Install `angr`, for example as follows:

        mkvirtualenv -p /usr/bin/python3 angr
        pip install angr
        pip install pyparsing
        workon angr

General usage in each `unrolldemo` directory:

        make

The `7-broadcast32` example requires `gcc -m32`, and intentionally
breaks `angr` to show what unsupported instructions look like.

The `10-ntt` example shows full functional verification (including range
verification and algebraic verification) for a fast asm NTT from 
[https://github.com/pq-crystals/kyber/blob/master/avx2/ntt.s](https://github.com/pq-crystals/kyber/blob/master/avx2/ntt.s).
The most interesting output files are

* `ntt.unroll`: output from the angr run
* `ntt.opt`: some peephole optimizations
* `ntt.range`: output of the range analyzer (including algebraic equations)
* `ntt.linear`: linear functions mod q

all of which are also included in the tarball for easier reference.

Tweaks to the NTT approach will sometimes require extra code in the
verifier, but a verifier that's only slightly larger than this should be
able to handle a much larger corpus of platform-optimized NTT
implementations. The point is that this reduces the large problem of
eliminating bugs in all those implementations to the much smaller
problem of eliminating bugs in the verifier.

### Notes on formal verification

What the NTT verifier is _meant_ to be doing is called "equivalence
checking", specifically checking that a short NTT program (`expntt`)
always produces the same outputs modulo q as the asm implementation. The
verifier is meant to be "sound": if it is sound and says that two
programs are equivalent then the programs are equivalent. The verifier
is not meant to be, and is not, "complete": it can fail to detect the
equivalence of equivalent programs. However, the verifier is
sufficiently complete to be able to handle this interesting example of
NTT code, and (I claim) many more examples.

_Formalizing_ equivalence requires specifying the languages that the
supposedly equivalent programs use, or at least large enough subsets of
the languages to handle the programs and the relevant input-output
behavior. Formalizing _soundness_ also requires specifying the language
that the verifier uses. Language specifications typically have errors
(compared to what was intended), but one can imagine a future in which
(1) the errors are fixed, (2) bugs in the language implementations are
fixed, and (3) the language implementations are verified to match the
specifications, assuming properties of the underlying hardware.
(In the meantime, _perhaps_ discrepancies between the language
specifications and the language implementations do not affect the
correctness of the verifier.)

_Proving_ soundness is obviously impossible if the verifier is not
sound, i.e., if the verifier has bugs. The verifier isn't doing anything
particularly tricky, so there are no obvious obstacles to a proof once
any bugs are fixed; but saying this is much easier than doing it,
especially if one requires the proof to be checked by a computer. Proofs
that aren't checked by a computer often have errors. Also,
[computerized proof checkers have had bugs](https://github.com/clarus/falso)
allowing bogus proofs. But one can imagine a future in which proof
checkers are themselves sound. (In the meantime, _perhaps_ bugs in the
proof checkers do not affect any proofs that we actually care about.)

From a security perspective, each layer here is creating another way to
let the computing base scale while the _trusted_ computing base stays
small. For example, instead of asking security reviewers to verify a
huge pile of asm implementations, we can have an equivalence checker
verify all the implementations, and merely ask security reviewers to
verify the equivalence checker. Even if the equivalence checker is
domain-specific (e.g., an equivalence checker only for linear functions
modulo small constants), this can be a huge win even for a small number
of asm implementations (depending on how simple the checker is), and
each new asm implementation makes the equivalence checker even more
attractive. As another example, instead of asking security reviewers to
verify many equivalence checkers, we can have a proof checker check
proofs of correctness of the equivalence checkers, and merely ask
security reviewers to verify the proof checker.
